package be.iminds.ilabt.jfed.espec.parser;

import be.iminds.ilabt.jfed.espec.model.*;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import org.jetbrains.annotations.Contract;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

/**
 * ExperimentSpecificationParser
 */
public class ExperimentSpecificationParser {
    public class ExperimentSpecificationParseException extends Exception {
        public ExperimentSpecificationParseException() { }
        public ExperimentSpecificationParseException(String s) { super(s); }
        public ExperimentSpecificationParseException(String s, Throwable throwable) { super(s, throwable); }
        public ExperimentSpecificationParseException(Throwable throwable) { super(throwable); }
        public ExperimentSpecificationParseException(String s, Throwable throwable, boolean b, boolean b1) { super(s, throwable, b, b1); }
    }

    private ObjectMapper mapper;
    public ExperimentSpecificationParser() {
        YAMLFactory yf = new YAMLFactory();
        this.mapper = new ObjectMapper(yf);
    }

    public ExperimentSpecification parse(String input) throws ExperimentSpecificationParseException {
        Map root;
        try {
            root = mapper.readValue(input, Map.class);
        } catch (IOException e) {
            throw new ExperimentSpecificationParseException("Failed to parse ExperimentSpecification yaml", e);
        }
        if (root == null)
            throw new ExperimentSpecificationParseException("Failed to parse ExperimentSpecification yaml");

        ExperimentSpecificationBuilder res = new BasicExperimentSpecificationBuilder();

        Object rawVersion = root.get("version");
        Object rawRspec = root.get("rspec");
        Object rawDir = root.get("dir");
        Object rawUpload = root.get("upload");
        Object rawExecute = root.get("execute");

        if (rawVersion == null)
            throw new ExperimentSpecificationParseException("Experiment Specification must contain version");
        if (!(rawVersion instanceof String))
            throw new ExperimentSpecificationParseException("Experiment Specification version must be a String");

        try {
            res.setVersion((String) rawVersion);
        } catch (IllegalArgumentException e) {
            throw new ExperimentSpecificationParseException("Experiment Specification version must be a String " +
                    "of the form <base>-<flavor>", e);
        }

        res.setRspec(extractRspecSpec(rawRspec));
        res.setDirs(extractDirSpecs(rawDir));
        res.setUploads(extractUploadSpecs(rawUpload));
        res.setExecutes(extractExecuteSpecs(rawExecute));

        //TODO: Make BasicExperimentSpecification the ExperimentSpecification, so no interface.
        //TODO    But make a factory/builder interface with getters and implement that with and without JavaFX
        return res.create();
    }

    @Nonnull
    private List<DirSpec> extractDirSpecs(@Nullable Object rawDir) throws ExperimentSpecificationParseException {
        if (rawDir == null)
            return Collections.emptyList();

        if (rawDir instanceof String) {
            return Collections.singletonList(new DirSpec((String) rawDir, null, null, null));
        } else {
            if (rawDir instanceof Map) {
                return Collections.singletonList(mapToDirSpec((Map)rawDir, "dir"));
            } else {
                if (rawDir instanceof List) {
                    List list = (List) rawDir;

                    List<DirSpec> res = new ArrayList<>();
                    for (int i = 0; i < list.size(); i++) {
                        Object listObj = list.get(i);

                        if (listObj instanceof String) {
                            res.add(new DirSpec((String) listObj, null, null, null));
                        } else {
                            if (listObj instanceof Map) {
                                res.add(mapToDirSpec((Map)listObj, "dir["+i+"]"));
                            } else {
                                throw new ExperimentSpecificationParseException("Experiment Specification dir list must not have item of type \"" + listObj.getClass().getName() + "\" (at index "+i+"))");
                            }
                        }
                    }
                    return res;
                } else {
                    throw new ExperimentSpecificationParseException("Experiment Specification dir must not be of type \""+rawDir.getClass().getName()+"\")");
                }
            }
        }
    }

    @Nonnull
    private List<UploadSpec> extractUploadSpecs(@Nullable Object rawUpload) throws ExperimentSpecificationParseException {
        if (rawUpload == null)
            return Collections.emptyList();

        if (rawUpload instanceof String) {
            return Collections.singletonList(new UploadSpec(new FileSource(FileSource.SourceType.BUNDLED, (String) rawUpload), null, null, null));
        } else {
            if (rawUpload instanceof Map) {
                return Collections.singletonList(mapToUploadSpec((Map)rawUpload, "upload"));
            } else {
                if (rawUpload instanceof List) {
                    List list = (List) rawUpload;

                    List<UploadSpec> res = new ArrayList<>();
                    for (int i = 0; i < list.size(); i++) {
                        Object listObj = list.get(i);

                        if (listObj instanceof String) {
                            res.add(new UploadSpec(new FileSource(FileSource.SourceType.BUNDLED, (String) listObj), null, null, null));
                        } else {
                            if (listObj instanceof Map) {
                                res.add(mapToUploadSpec((Map)listObj, "upload["+i+"]"));
                            } else {
                                throw new ExperimentSpecificationParseException("Experiment Specification upload list must not have item of type \"" + listObj.getClass().getName() + "\" (at index "+i+"))");
                            }
                        }
                    }
                    return res;
                } else {
                    throw new ExperimentSpecificationParseException("Experiment Specification upload must not be of type \""+rawUpload.getClass().getName()+"\")");
                }
            }
        }
    }

    @Nonnull
    private List<ExecuteSpec> extractExecuteSpecs(@Nullable Object rawExecute) throws ExperimentSpecificationParseException {
        if (rawExecute == null)
            return Collections.emptyList();

        if (rawExecute instanceof String) {
            return Collections.singletonList(new ExecuteSpec(new FileSource(FileSource.SourceType.BUNDLED, (String) rawExecute), null, null, null, null));
        } else {
            if (rawExecute instanceof Map) {
                return Collections.singletonList(mapToExecuteSpec((Map)rawExecute, "execute"));
            } else {
                if (rawExecute instanceof List) {
                    List list = (List) rawExecute;

                    List<ExecuteSpec> res = new ArrayList<>();
                    for (int i = 0; i < list.size(); i++) {
                        Object listObj = list.get(i);

                        if (listObj instanceof String) {
                            res.add(new ExecuteSpec(new FileSource(FileSource.SourceType.BUNDLED, (String) listObj), null, null, null, null));
                        } else {
                            if (listObj instanceof Map) {
                                res.add(mapToExecuteSpec((Map)listObj, "execute["+i+"]"));
                            } else {
                                throw new ExperimentSpecificationParseException("Experiment Specification execute list must not have item of type \"" + listObj.getClass().getName() + "\" (at index "+i+"))");
                            }
                        }
                    }
                    return res;
                } else {
                    throw new ExperimentSpecificationParseException("Experiment Specification execute must not be of type \""+rawExecute.getClass().getName()+"\")");
                }
            }
        }
    }

    @Nonnull
    private DirSpec mapToDirSpec(@Nonnull Map in, @Nonnull String location) throws ExperimentSpecificationParseException {
        String path = extractNullableStringByKey(in, "path", location);
        if (path == null)
            throw new ExperimentSpecificationParseException("Experiment Specification "+location+" must have path)");
        String content = extractNullableStringByKey(in, "content", location);
        String permissions = extractNullableStringByKey(in, "permission", location);
        List<String> nodes = objectToNullableStringList(in.get("nodes"), location+".nodes");
        return new DirSpec(path, content, permissions, nodes);
    }


    @Nonnull
    private UploadSpec mapToUploadSpec(@Nonnull Map in, @Nonnull String location) throws ExperimentSpecificationParseException {
        FileSource source = mapToFileSource(in, location);
        String path = extractNullableStringByKey(in, "path", location);
        String permissions = extractNullableStringByKey(in, "permission", location);
        List<String> nodes = objectToNullableStringList(in.get("nodes"), location+".nodes");
        return new UploadSpec(source, path, permissions, nodes);
    }


    @Nonnull
    private ExecuteSpec mapToExecuteSpec(@Nonnull Map in, @Nonnull String location) throws ExperimentSpecificationParseException {
        FileSource source = mapToFileSource(in, location);
        String path = extractNullableStringByKey(in, "path", location);
        String pwd = extractNullableStringByKey(in, "pwd", location);
        String permissions = extractNullableStringByKey(in, "permission", location);
        List<String> nodes = objectToNullableStringList(in.get("nodes"), location+".nodes");
        return new ExecuteSpec(source, path, pwd, permissions, nodes);
    }

    @Nullable
    @Contract("null,_ -> null; !null,_ -> !null")
    private List<String> objectToNullableStringList(@Nullable Object in, @Nonnull String location) throws ExperimentSpecificationParseException {
        if (in == null)
            return null;

        if (in instanceof String) {
            return Collections.singletonList((String) in);
        } else {
            if (in instanceof List) {
                List list = (List) in;
                List<String> res = new ArrayList<>(list.size());
                for (int i = 0; i < list.size(); i++) {
                    Object listItemObj = list.get(i);
                    if (listItemObj instanceof String) {
                        res.add((String) listItemObj);
                    } else
                        throw new ExperimentSpecificationParseException("Experiment Specification expect type String instead of \""+in.getClass().getName()+"\" at "+location+"["+i+"]");
                }
                return res;
            } else {
                throw new ExperimentSpecificationParseException("Experiment Specification did not expect type \""+in.getClass().getName()+"\" at "+location);
            }
        }
    }

    @Nullable
    private String extractNullableStringByKey(@Nonnull Map m, @Nonnull String key, @Nonnull String location) throws ExperimentSpecificationParseException {
        Object in = m.get(key);
        if (in == null)
            return null;

        if (in instanceof String) {
            return (String) in;
        } else {
            throw new ExperimentSpecificationParseException("Experiment Specification did not expect type \""+in.getClass().getName()+"\" at "+location+"."+key);
        }
    }

    @Nullable
    @Contract("null,_ -> null; !null,_ -> !null")
    private String objectToString(@Nullable Object in, @Nonnull String location) throws ExperimentSpecificationParseException {
        if (in == null)
            return null;

        if (in instanceof String) {
            return (String) in;
        } else {
            throw new ExperimentSpecificationParseException("Experiment Specification did not expect type \""+in.getClass().getName()+"\" at "+location);
        }
    }

    @Nonnull
    private RspecSpec extractRspecSpec(@Nullable Object rawRspec) throws ExperimentSpecificationParseException {
        if (rawRspec == null)
            throw new ExperimentSpecificationParseException("Experiment Specification must contain rspec");

        if (rawRspec instanceof String) {
            return new RspecSpec(new FileSource(FileSource.SourceType.BUNDLED, (String) rawRspec));
        } else {
            if (rawRspec instanceof Map) {
                return new RspecSpec(mapToFileSource((Map)rawRspec, "rspec"));
            } else {
                if (rawRspec instanceof List) {
                    List rspecList = (List) rawRspec;

                    if (rspecList.size() != 1)
                        throw new ExperimentSpecificationParseException("Experiment Specification rspec must not be list of size 1, not of size "+rspecList.size()+")");

                    Object listObj = rspecList.get(0);
//                    for (int i = 0; i < rspecList.size(); i++) {
//                        Object listObj = rspecList.get(i);

                    if (listObj instanceof String) {
                        return new RspecSpec(new FileSource(FileSource.SourceType.BUNDLED, (String) listObj));
                    } else {
                        if (listObj instanceof Map) {
                            return new RspecSpec(mapToFileSource((Map)listObj, "rspec"));
                        } else {
                            throw new ExperimentSpecificationParseException("Experiment Specification rspec list must not have item of type \""+listObj.getClass().getName()+"\")");
                        }
                    }
                } else {
                    throw new ExperimentSpecificationParseException("Experiment Specification rspec must not be of type \""+rawRspec.getClass().getName()+"\")");
                }
            }
        }
    }


    @Nonnull
    private FileSource mapToFileSource(@Nonnull Map in, String location) throws ExperimentSpecificationParseException {
        for (FileSource.SourceType sourceType : FileSource.SourceType.values()) {
            String key = sourceType.name().toLowerCase();
            Object val = in.get(key);
            if (val != null) {
                if (val instanceof String) {
                    return new FileSource(sourceType, (String) val);
                } else {
                    throw new ExperimentSpecificationParseException("Illegal entry: " + key + ": " + val.getClass().getName());
                }
            }
        }
        throw new ExperimentSpecificationParseException("No \"source\" (bundled, direct, ...) found in "+location);
    }
}
